import json
from typing import Union
from dataclasses import dataclass

from dacite import from_dict


@dataclass
class CommonData:
    maxEpochMinutes: int
    nodeBlockTimeWindow: int
    perRoundBlocks: int
    maxConsensusVals: int
    additionalCycleTime: int


@dataclass
class StakingData:
    stakeThreshold: int
    operatingThreshold: int
    maxValidators: int
    unStakeFreezeDuration: int
    rewardPerMaxChangeRange: int
    rewardPerChangeInterval: int


@dataclass
class SlashingData:
    slashFractionDuplicateSign: int
    duplicateSignReportReward: int
    maxEvidenceAge: int
    slashBlocksReward: int
    zeroProduceCumulativeTime: int
    zeroProduceNumberThreshold: int
    zeroProduceFreezeDuration: int


@dataclass
class GovernData:
    versionProposalVoteDurationSeconds: int
    versionProposalSupportRate: int
    textProposalVoteDurationSeconds: int
    textProposalVoteRate: int
    textProposalSupportRate: int
    cancelProposalVoteRate: int
    cancelProposalSupportRate: int
    paramProposalVoteDurationSeconds: int
    paramProposalVoteRate: int
    paramProposalSupportRate: int


@dataclass
class RewardData:
    newBlockRate: int
    platonFoundationYear: int
    increaseIssuanceRatio: int
    theNumberOfDelegationsReward: int


@dataclass
class RestrictingData:
    minimumRelease: int


@dataclass
class Economic:
    common: CommonData
    restricting: RestrictingData
    staking: StakingData
    gov: GovernData
    reward: RewardData
    slashing: SlashingData

    # 区块
    @property
    def block_time(self):
        """ 区块时长/s
        """
        # todo: 获取区块平均时间
        return int(self.common.nodeBlockTimeWindow // self.common.perRoundBlocks)

    # 窗口期
    @property
    def round_time(self):
        """ 窗口期时长/m
        """
        return self.round_blocks * self.block_time

    @property
    def round_blocks(self):
        """ 窗口期区块数量
        """
        return self.common.perRoundBlocks

    # 共识轮
    @property
    def consensus_time(self):
        """ 共识轮时长/m
        """
        return self.consensus_rounds * self.round_time

    @property
    def consensus_blocks(self):
        """ 共识轮区块数量
        """
        return self.consensus_rounds * self.round_blocks

    @property
    def consensus_rounds(self):
        """ 共识轮窗口期数量
        """
        return self.validator_count

    # 结算周期
    @property
    def epoch_time(self):
        """ 结算周期时长/m
        """
        return self.epoch_consensus * self.consensus_time

    @property
    def epoch_blocks(self):
        """ 结算周期区块数量
        """
        return self.epoch_consensus * self.consensus_blocks

    @property
    def epoch_rounds(self):
        """ 结算周期窗口期数量
        """
        return self.epoch_consensus * self.consensus_rounds

    @property
    def epoch_consensus(self):
        """ 结算周期共识轮数量
        """
        return (self.common.maxEpochMinutes * 60) // self.consensus_time

    @property
    def increasing_time(self):
        """ 增发周期时长/m
        """
        return self.increasing_epoch * self.epoch_time

    @property
    def increasing_blocks(self):
        """ 增发周期区块数
        """
        return self.increasing_rounds * self.round_blocks

    @property
    def increasing_rounds(self):
        """ 增发周期窗口期数
        """
        return self.increasing_consensus * self.consensus_rounds

    @property
    def increasing_consensus(self):
        """ 增发周期共识轮数
        """
        return self.increasing_epoch * self.epoch_consensus

    @property
    def increasing_epoch(self):
        """ 增发周期结算周期数
        """
        # todo: 修改为实时计算
        return (self.common.additionalCycleTime * 60) // self.epoch_time

    @property
    def verifier_count(self):
        """ 最大共识验证人数量
        """
        return self.common.maxConsensusVals

    @property
    def validator_count(self):
        """ 最大验证人数量
        """
        return self.staking.maxValidators

    @property
    def staking_limit(self):
        """ 质押最小金额限制
        """
        return self.staking.stakeThreshold

    @property
    def add_staking_limit(self):
        """ 增持质押最小金额限制
        """
        return self.staking.operatingThreshold

    @property
    def delegate_limit(self):
        """ 委托最小金额限制
        """
        return self.staking.operatingThreshold

    @property
    def unstaking_freeze_epochs(self):
        """ 解质押后，质押金额冻结的结算周期数
        """
        return self.staking.unStakeFreezeDuration

    @property
    def not_block_slash_rate(self):
        """ 节点零出块处罚时，罚金对应的区块奖励倍数
        """
        return self.slashing.slashBlocksReward

    @property
    def param_proposal_epochs(self):
        """ 参数提案投票期的结算周期数
        """
        return self.govern.paramProposalVoteDurationSeconds // self.epoch_time

    @property
    def text_proposal_epochs(self):
        """ 文本提案投票期的结算周期数
        """
        return self.govern.textProposalVoteDurationSeconds // self.epoch_time


def new_economic(data: Union[dict, str]):
    """ 将data转换为economic对象
    """
    if type(data) is str:
        data = json.loads(data)

    return from_dict(Economic, data)



import functools
import json
import os
import sys
from os.path import abspath
from typing import cast

import rlp
from hexbytes import HexBytes

from platon import Web3, HTTPProvider, WebsocketProvider, IPCProvider
from platon._utils.inner_contract import InnerContractEvent
from platon._utils.threads import Timeout
from platon.datastructures import AttributeDict
from platon.exceptions import ContractLogicError
from platon.types import BlockData, CodeData
from platon_account._utils.signing import to_standard_signature_bytes
from platon_hash.auto import keccak
from platon_keys.datatypes import Signature
from platon_typing import HexStr
from platon_utils import remove_0x_prefix, to_canonical_address

from gql import Client
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.websockets import WebsocketsTransport


def get_web3(uri, chain_id=None, hrp=None, timeout=10, modules=None):
    """ 通过rpc uri，获取web3对象。可以兼容历史platon版本
    """
    if uri.startswith('http'):
        provider = HTTPProvider
    elif uri.startswith('ws'):
        provider = WebsocketProvider
    elif uri.startswith('ipc'):
        provider = IPCProvider
    else:
        raise ValueError(f'unidentifiable uri {uri}')

    with Timeout(timeout) as t:
        while True:
            web3 = Web3(provider(uri), chain_id=chain_id, hrp=hrp, modules=modules)
            if web3.isConnected():
                break
            t.sleep(2)

    return web3


def get_gql(uri):
    """ 通过gql uri，获取gql对象。
    # todo: 增加超时处理
    """
    if uri.startswith('http'):
        transport = AIOHTTPTransport
    elif uri.startswith('ws'):
        transport = WebsocketsTransport
    else:
        raise ValueError(f'unidentifiable uri {uri}')

    return Client(transport=transport(uri), fetch_schema_from_transport=True)


def send_transaction(web3: Web3, txn: dict, private_key: str):
    """
    签名交易并发送，可以获取交易hash或交易回执

    Args:
        web3: Web3对象
        txn: 要发送的交易dict
        private_key: 地址私钥，用于签名交易
        returns: 指定要返回的结果，取值如下：
                - 'hash': 返回交易哈希
                - 'receipt': 返回交易回执
                - 'event': 返回内置合约的event内容
    """
    if not private_key:
        return web3.platon.send_transaction(txn)

    if not txn.get('nonce'):
        account = web3.platon.account.from_key(private_key, hrp=web3.hrp)
        txn['nonce'] = web3.platon.get_transaction_count(account.address)

    signed_txn = web3.platon.account.sign_transaction(txn, private_key, web3.hrp)
    return web3.platon.send_raw_transaction(signed_txn.rawTransaction)


def get_transaction_result(web3: Web3, tx_hash, result_type):
    """ 根据指定的result type，来获取交易的返回值
    """
    if result_type == 'hash':
        return bytes(tx_hash).hex()
    receipt = web3.platon.wait_for_transaction_receipt(tx_hash)
    if type(receipt) is bytes:
        receipt = receipt.decode('utf-8')
    if result_type == 'receipt':
        return receipt
    raise ValueError(f'unknown result type: {result_type}')


def contract_call(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        return func(self, *args, **kwargs).call()

    return wrapper


def contract_transaction(func_id=None):
    """ todo: 增加注释
    """

    def out_wrapper(func):

        @functools.wraps(func)
        def wrapper(self, *args, txn=None, private_key=None, **kwargs):
            # 预填充from地址，避免预估gas时地址相关检验不通过
            account = self.web3.platon.account.from_key(private_key, hrp=self.web3.hrp) if private_key else self.default_account
            if not txn:
                txn = {}
            if not txn.get('from'):
                txn['from'] = account.address

            if func.__name__ == 'fit_func':
                # solidity合约方法不传入private key参数，避免abi解析问题
                fn = func(self, *args, **kwargs)
            else:
                # 内置合约
                fn = func(self, *args, private_key=private_key, **kwargs)

            # 预估gas出现合约逻辑错误时，不再报错，而是返回data信息
            # 在txn中指定gas，可以跳过该步骤（避免出现太多种返回值）
            try:
                txn = fn.build_transaction(txn)
            except ContractLogicError as e:
                err = str(e)
                if err.startswith('inner contract exec failed: '):
                    event = err.split('inner contract exec failed: ')[1]
                    data = json.loads(event.replace('\'', '"'))
                    return cast(CodeData, AttributeDict.recursive(data))
                return err

            # 直接返回txn
            if self._result_type == 'txn':
                return txn

            if self._result_type == 'receipt' or self._result_type == 'hash':
                return self.send_transaction(txn, private_key, self._result_type)

            if self._result_type == 'event':
                if self._module_type != 'inner-contract':
                    raise TypeError('result type "event" only support inner contract')

                receipt = self.send_transaction(txn, private_key, 'receipt')
                return InnerContractEvent(func_id).processReceipt(receipt)

            raise ValueError(f'unknown result type: {self._result_type}')

        return wrapper

    return out_wrapper


def ec_recover(block: BlockData):
    """ 使用keccak方式，解出区块的签名节点公钥
    """
    extra = block.proofOfAuthorityData[:32]
    sign = block.proofOfAuthorityData[32:]
    raw_data = [bytes.fromhex(remove_0x_prefix(block.parentHash.hex())),
                to_canonical_address(block.miner),
                bytes.fromhex(remove_0x_prefix(block.stateRoot.hex())),
                bytes.fromhex(remove_0x_prefix(block.transactionsRoot.hex())),
                bytes.fromhex(remove_0x_prefix(block.receiptsRoot.hex())),
                bytes.fromhex(remove_0x_prefix(block.logsBloom.hex())),
                block.number,
                block.gasLimit,
                block.gasUsed,
                block.timestamp,
                extra,
                bytes.fromhex(remove_0x_prefix(block.nonce.hex()))
                ]
    hash_bytes = HexBytes(keccak(rlp.encode(raw_data)))
    signature_bytes = HexBytes(sign)
    signature_bytes_standard = to_standard_signature_bytes(signature_bytes)
    signature = Signature(signature_bytes=signature_bytes_standard)
    return remove_0x_prefix(HexStr(signature.recover_public_key_from_msg_hash(hash_bytes).to_hex()))


def run(cmd):
    """
    The machine executes the cmd command and gets the result
    :param cmd:
    :return:
    """
    r = os.popen(cmd)
    out = r.read()
    r.close()
    return out


def mock_duplicate_sign(dtype, sk, blskey, block_number, epoch=0, view_number=0, block_index=0, index=0):
    """
    forged double sign
    :param dtype:
    :param sk:
    :param blskey:
    :param block_number:
    :param epoch:
    :param view_number:
    :param block_index:
    :param index:
    :return:
    """
    if sys.platform in "linux,linux2":
        tool_file = abspath("tool/linux/duplicateSign")
        run("chmod +x {}".format(tool_file))
    else:
        tool_file = abspath("tool/win/duplicateSign.exe")
    print("{} -dtype={} -sk={} -blskey={} -blockNumber={} -epoch={} -viewNumber={} -blockIndex={} -vindex={}".format(
        tool_file, dtype, sk, blskey, block_number, epoch, view_number, block_index, index))
    output = run(
        "{} -dtype={} -sk={} -blskey={} -blockNumber={} -epoch={} -viewNumber={} -blockIndex={} -vindex={}".format(
            tool_file, dtype, sk, blskey, block_number, epoch, view_number, block_index, index))
    print(output)
    if not output:
        raise Exception("unable to use double sign tool")
    return output.strip("\n")



from gql import gql

from platon_aide.utils import get_gql


class Graphql:

    def __init__(self, uri: str):
        self.client = get_gql(uri)

    def execute(self, content):
        return self.client.execute(gql(content))



import time

from loguru import logger
from platon.main import get_default_modules
from platon.middleware import gplaton_poa_middleware
from platon_account import Account, DEFAULT_HRP
from platon_utils import to_bech32_address, to_checksum_address, combomethod
from platon_aide.transfer import Transfer
from platon_aide.restricting import Restricting
from platon_aide.economic import new_economic, Economic
from platon_aide.calculator import Calculator
from platon_aide.contract import Contract
from platon_aide.staking import Staking
from platon_aide.delegate import Delegate
from platon_aide.slashing import Slashing
from platon_aide.govern import Govern
from platon_aide.graphqls import Graphql
from platon_aide.utils import get_web3, ec_recover


def get_modules(exclude: list = None):
    """ 排除节点关闭的API
    """
    if not exclude:
        exclude = []

    modules = get_default_modules()
    if 'admin' in exclude:
        modules['node'][1].pop('admin')
    if 'debug' in exclude:
        modules.pop('debug')

    return modules


class Aide:
    """ 主类，platon各个子模块的集合体，同时支持创建账户、解码等非交易类的操作
    """
    hrp: str = ''

    def __init__(self,
                 uri: str,
                 gql_uri: str = None,
                 chain_id: int = None,
                 hrp: str = None,
                 economic: Economic = None,
                 exclude_api: list = None,
                 ):
        self.uri = uri
        self.gql_uri = gql_uri
        # web3相关设置
        modules = get_modules(exclude_api)
        self.web3 = get_web3(uri, chain_id, hrp, modules=modules)
        self.web3.middleware_onion.inject(gplaton_poa_middleware, layer=0)
        self.hrp = hrp or self.web3.hrp
        self.chain_id = chain_id or self.web3.platon.chain_id
        self.__set_web3_attr()
        # 属性对象设置
        self.economic = new_economic(self.debug.economic_config()) if self.debug else economic
        self.calculator = Calculator(self.web3, economic=self.economic)
        self.transfer = Transfer(self.web3)
        self.restricting = Restricting(self.web3)
        self.staking = Staking(self.web3, economic=self.economic)
        self.delegate = Delegate(self.web3, economic=self.economic)
        self.slashing = Slashing(self.web3)
        self.govern = Govern(self.web3)
        self.contract = Contract(self.web3)
        self.graphql = Graphql(self.gql_uri) if self.gql_uri else None

    def __set_web3_attr(self):
        self.platon = self.web3.platon
        self.admin = self.web3.node.admin if hasattr(self.web3, 'admin') else None
        self.txpool = self.web3.node.txpool if hasattr(self.web3.node, 'txpool') else None
        self.personal = self.web3.node.personal if hasattr(self.web3.node, 'personal') else None
        self.debug = self.web3.debug if hasattr(self.web3, 'debug') else None

    def set_default_account(self, account):
        """ 设置默认账户
        """
        self.transfer.set_default_account(account)
        self.restricting.set_default_account(account)
        self.staking.set_default_account(account)
        self.delegate.set_default_account(account)
        self.slashing.set_default_account(account)
        self.govern.set_default_account(account)
        self.contract.set_default_account(account)

    def set_result_type(self, result_type):
        self.transfer.set_result_type(result_type)
        self.restricting.set_result_type(result_type)
        self.staking.set_result_type(result_type)
        self.delegate.set_result_type(result_type)
        self.slashing.set_result_type(result_type)
        self.govern.set_result_type(result_type)
        self.contract.set_result_type(result_type)

    @combomethod
    def create_account(self, hrp=None):
        """ 创建账户
        """
        hrp = hrp or self.hrp or DEFAULT_HRP
        return Account.create(hrp=hrp)

    @combomethod
    def create_hd_account(self, hrp=None):
        """ 创建HD账户
        """
        pass

    @combomethod
    def create_keystore(self, passphrase, key=None):
        """ 创建钱包文件
        """
        pass

    @combomethod
    def to_bech32_address(self, address, hrp=None):
        """ 任意地址转换为platon形式的bech32地址
        注意：非标准bech32地址
        """
        hrp = hrp or self.hrp or DEFAULT_HRP
        return to_bech32_address(address, hrp=hrp)

    @combomethod
    def to_checksum_address(self, address):
        """ 任意地址转换为checksum地址
        """
        return to_checksum_address(address)

    def wait_block(self, to_block, time_out=None):
        """ 等待块高
        """
        current_block = self.platon.block_number
        time_out = time_out or (to_block - current_block) * 3

        for i in range(time_out):
            time.sleep(1)
            current_block = self.platon.block_number

            if i % 10 == 0:
                logger.info(f'waiting block: {current_block} -> {to_block}')

            if current_block >= to_block:
                return

        raise TimeoutError('wait block timeout!')

    def ec_recover(self, block_identifier):
        """ 获取出块节点公钥
        """
        block = self.web3.platon.get_block(block_identifier)
        return ec_recover(block)
